
//
// main.c
//
// <See class's '.h' file>
//
// $Id: main.c,v 1.23 2001/04/23 23:54:55 dru Exp $
//
//

#include "svm32.h"

#include "celendian.h"
#include "atom.h"
#include "runtime.h"
#include "assemble.h"
#include "compiler.h"
#include "cel2sd.h"
#include "string.h"
#include "array.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


int debugParse;
int debugParseLevel = -1;
int debugCodeGen;
int debugForce;
extern char **environ;


int main (int argc, char *argv[]) 
{
    
    unsigned char * p;
    unsigned char * filename;
    int i;
    int needsCompile;

    Proto module;
    Proto environment;
    Proto arguments;
    void * mem;
    int length;
    char buff[4096];
    int sawFilename = 0;
    char baseName[256];
    
    

    VM32Cpu cpu;
    VM32Cpu * pcpu;
    
    
    // Setup the memory and the GC
    

    // Setup our virtual CPU
    pcpu = &cpu;
    init(pcpu);
    reset(pcpu);


    // Setup our runtime
    celInit(0, pcpu);

    p = UniqueString("car");
    p = UniqueString("car");
    
    for (i = 1; i < argc; i++) {
       if (argv[i][0] == '-') {
         if (strcmp(argv[i], "--trace") == 0) {
	   debugTrace = 1;
           printf("debugTrace ON\n");
         } else 
         if (strcmp(argv[i], "--dump") == 0) {
	   debugDump = 1;
           printf("debugDump ON\n");
         } else
         if (strcmp(argv[i], "--force") == 0) {
	   debugForce = 1;
           printf("debugForce ON\n");
         } else
         if (strcmp(argv[i], "--parse") == 0) {
	   debugParse = 1;
           printf("debugParse ON\n");
         } else
         if (strcmp(argv[i], "--parseLevel") == 0) {
	     i++;
	     if (i > argc) {
		 printf("Missing argument to parseLevel\n");
		 exit(1);
	     }
	     if (!isdigit(argv[i][0])) {
		 printf("Improper argument to parseLevel\n");
		 exit(1);
	     }
	     debugParseLevel = atoi(argv[i]);
	     printf("debugParseLevel = %d ON\n", debugParseLevel);
         } else
         if (strcmp(argv[i], "--codegen") == 0) {
	   debugCodeGen = 1;
           printf("debugCodeGen ON\n");
         }
       } else {
           if (sawFilename != 1) {
	     filename = argv[i];
	     sawFilename = 1;
           }
       }
    }

    // Load the module
    if (argc == 1) {
	printf("Syntax: cel --parse --parseLevel <num> --codegen --trace --dump <module file>\n\n");
	printf("  parse:      Display parse tree\n");
	printf("  parseLevel: Display this level x (0=all)\n");
	printf("  codegen:    Display the generated code\n");
	printf("  trace:      Show a trace of the running code\n");
	printf("  dump:       Display the pre-compiled code module (if exists) \n");

	exit(1);
    }

    // If filename has '.cel' suffix, then remove that
    p = strstr(filename, ".cel");
    if (NULL != p) {
      // If the suffix is present, remove it
      *p = 0;
    }

    // If filename has a '/' then copy all of the filename
    // up to the last '/' as the basename
    p = celRevStrStr(filename, strlen(filename), "/", 1);
    if (p != NULL) {
	memcpy(baseName, filename, (p - filename));
	baseName[p-filename] = 0;
	filename = p + 1;
	
    } else {
	baseName[0] = 0;
    }


    // Check if the module needs to be compiled to disk
    needsCompile = checkCompiled(filename);


    if (needsCompile || debugForce) {
	// Load the module
	module = compile(loadModuleStream(filename), debugParse, debugCodeGen);
	mem = cel2sd(module, &length);
	if (writeModule(filename, mem, length)) {
	  printf("Couldn't write compiled module [%s]\n\n", filename);
	  exit(1);
	} 
    }
    

    // Prepare the module for running
    // args(filename, donotRunflag) 
    pcpu = doCelInit(filename, 0);

    // Setup the Environment
    i = 0;

    environment = objectNew(0);
    arguments = createArray(nil);
    objectSetSlot(Capsule, stringToAtom("environment"), PROTO_ACTION_GET, (uint32) environment);
    objectSetSlot(Capsule, stringToAtom("arguments"), PROTO_ACTION_GET, (uint32) arguments);
    
    while (1) {
	if (environ[i] == NULL) {
	    break;
	}

	strcpy(buff, environ[i]);
	
	p = strchr(buff, '=');
	*p= 0;
	p++;
	
	objectSetSlot(environment, stringToAtom(buff), PROTO_ACTION_GET, (uint32) stringToAtom(p));
	i++;
    }
    
    for (i = 0; i < argc; i++) {
	//printf("%d - %s\n", i, argv[i]);
	addToArray(arguments, (Proto) stringToAtom(argv[i]));
    }

    //objectDump(arguments);
    
    
    if (debugTrace) {
      // Turn on debug mode
      runtimeActivateTrace = 1;
    }



    // Run the code straight
    run (pcpu);
    
    return 0;
}

